from typing import Tuple, List

from postgrest.exceptions import APIError
import numpy as np

from tensorage.types import Dataset
from .base import BaseContext



class DatabaseContext(BaseContext):
    def __setup_auth(self):
        # store the current JWT token
        self._anon_key = self.backend.client.supabase_key

        # set the JWT of the authenticated user as the new token
        self.backend.client.postgrest.auth(self.backend._session.access_token)
    
    def __restore_auth(self):
        # restore the original JWT
        self.backend.client.postgrest.auth(self._anon_key)

    def check_schema_installed(self) -> bool:
        # setup auth token
        self.__setup_auth()

        # check if the datasets and tensor_float4 tables exist
        missing_table = False

        for table in ('datasets', 'tensors_float4'):
            try:
                self.backend.client.table(table).select('*', count='exact').limit(1).execute()
            except APIError as e:
                if e.code == '42P01':
                    missing_table = True
                else:
                    raise e
        
        # check if any of the needed tables was not found
        return not missing_table

    def insert_dataset(self, key: str, shape: Tuple[int], dim: int) -> Dataset:
        # run the insert
        self.__setup_auth()
        response = self.backend.client.table('datasets').insert({'key': key, 'shape': shape, 'ndim': dim, 'user_id': self.user_id}).execute()
        self.__restore_auth()

        # return an instance of Dataset
        data = response.data[0]
        return Dataset(id=data['id'], key=data['key'], shape=data['shape'], ndim=data['ndim'])
    
    def insert_tensor(self, data_id: int, data: List[np.ndarray], offset: int = 0) -> bool:
        # setup auth token
        self.__setup_auth()

        # run the insert
        try:
            self.backend.client.table('tensors_float4').insert([{'data_id': data_id, 'index': int(i + 1 + offset), 'user_id': self.user_id, 'tensor': chunk.tolist()} for i, chunk in enumerate(data)]).execute()
        except APIError as e:
            # TODO check if we expired here and refresh the token
            raise e
        
        # restore old token
        self.__restore_auth()

        # return 
        return True

    def get_dataset(self, key: str) -> Dataset:
        # setup auth token
        self.__setup_auth()

        # get the dataset
        response = self.backend.client.table('datasets').select('*').eq('key', key).execute()

        # restore old token
        self.__restore_auth()

        # grab the data
        data = response.data[0]

        # return as Dataset
        return Dataset(id=data['id'], key=data['key'], shape=data['shape'], ndim=data['ndim'])

    def get_tensor(self, key: str, index_low: int, index_up: int, slice_low: List[int], slice_up: List[int]) -> np.ndarray:
        # setup auth token
        self.__setup_auth()

        # get the requested chunk
        response = self.backend.client.rpc('tensor_float4_slice', {'name': key, 'index_low': index_low, 'index_up': index_up, 'slice_low': slice_low, 'slice_up': slice_up}).execute()

        # restore old token
        self.__restore_auth()

        # grab the data
        data = response.data[0]['tensor']

        # return as np.ndarray
        return np.asarray(data)

    def remove_dataset(self, key: str) -> bool:
        # setup auth token
        self.__setup_auth()

        # remove the dataset
        self.backend.client.table('datasets').delete().eq('key', key).execute()

        # restore old token
        self.__restore_auth()
        
        # return
        return True

    def list_dataset_keys(self) -> List[str]:
        # setup auth token
        self.__setup_auth()

        # get the keys
        response = self.backend.client.table('datasets').select('key').execute()

        # restore old token
        self.__restore_auth()

        return [row['key'] for row in response.data]

    def append_tensor(self, data_id: int, data: List[np.ndarray]) -> bool:
        return super().append_tensor(data_id, data)



from typing import Any, TypeVar, Generic
from dataclasses import dataclass, field

from supabase import Client, create_client
from gotrue.types import AuthResponse, User, Session
from dotenv import load_dotenv

from .store import TensorStore
from .backend.base import BaseContext
from .backend.database import DatabaseContext
from .backend.storage import StorageContext

load_dotenv()


C = TypeVar('C', bound=BaseContext)

@dataclass
class ContextWrapper(Generic[C]):
    _session: 'BackendSession'
    Context: type[C]

    def __enter__(self) -> C:
        # login the session if it is not logged in
        if not hasattr(self._session, '_session') or self._session._session is None:
            self._session.login_by_mail()
        
        # instatiate the store with an authenticated Session
        context = self.Context(self._session)

        return context

    def __exit__(self, *args):
        # logout the session
        self._session.logout()


@dataclass
class BackendSession(object):
    email: str
    password: str
    backend_url: str 
    backend_key: str = field(repr=False)
    _client: Client = field(init=False, repr=False)
    _user: User = field(init=False, repr=False)
    _session: Session = field(init=False, repr=False)


    @property
    def client(self) -> Client:
        if not hasattr(self, '_client') or self._client is None:
            self._client = create_client(self.backend_url, self.backend_key)
        return self._client

    def login_by_mail(self) -> AuthResponse:
        # login
        response = self.client.auth.sign_in_with_password({'email': self.email, 'password': self.password})

        # store user and session info
        self._user = response.user
        self._session = response.session
        # return response
        return response
    
    def register_by_mail(self, email: str, password: str) -> AuthResponse:
        # register
        response = self.client.auth.sign_up({'email': email, 'password': password})
        
        # store user and session info
        self._user = response.user
        self._session = response.session

        # return response
        return response

    def refresh(self) -> AuthResponse:
        # refresh
        response = self.client.auth.refresh_session(self._session.refresh_token)

        # renew tokens
        self._session = response.session

        # return response
        return response
    
    def logout(self):
        if hasattr(self, '_client') and self._client is not None:
            self.client.auth.sign_out()

    def database(self) -> ContextWrapper[DatabaseContext]:
        return ContextWrapper(self, DatabaseContext)
    
    def storage(self) -> ContextWrapper[StorageContext]:
        return ContextWrapper(self, StorageContext)

    def __del__(self):
        self.logout()

    def __call__(self) -> TensorStore:
        # init a store
        return TensorStore(self)



from typing import List

from .base import BaseContext

class StorageContext(BaseContext):
    def __setup_auth(self):
        # store the current JWT token
        self._anon_key = self.backend.client.supabase_key

        # add the authenticated JWT to the headers
        self.backend.client.storage._client.headers['Authorization'] = f"Bearer {self.backend._session.access_token}"

    def __restore_auth(self):
        # restore the original JWT
        self.backend.client.storage._client.headers['Authorization'] = f"Bearer {self._anon_key}"

    def _create_user_bucket(self):
        raise NotImplementedError

    def has_bucket(self) -> bool:
        raise NotImplementedError

    def list_datasets(self) -> List[str]:
        raise NotImplementedError


from typing import TYPE_CHECKING, Tuple, Union, List
from dataclasses import dataclass, field
import warnings

from tqdm import tqdm
import numpy as np

if TYPE_CHECKING:
    from tensorage.session import BackendSession


@dataclass
class TensorStore(object):
    backend: 'BackendSession' = field(repr=False)
    quiet: bool = field(default=False)

    # some stuff for upload
    chunk_size: int = field(default=100000, repr=False)

    def __post_init__(self):
        # check if the schema is installed
        with self.backend.database() as db:
            if not db.check_schema_installed():
                from tensorage.db_init import SQL
                warnings.warn(f"The schema for the TensorStore is not installed. Please connect the database and run the following script:\n\n--------8<--------\n{SQL}\n\n--------8<--------\n")

    def __getitem__(self, key: Union[str, Tuple[Union[str, slice, int]]]):
        # first get key
        if isinstance(key, str):
            name = key
        elif isinstance(key[0], str):
            name = key[0]
        else:
            raise KeyError('You need to pass the key as first argument.')
        
        # load the dataset
        with self.backend.database() as db:
            dataset = db.get_dataset(name)

        # now we need to figure out, what kind of slice we need to pass
        if isinstance(key, str):
            index = [1, dataset.shape[0] + 1]
            slices = [[1, dataset.shape[i] + 1] for i in range(1, dataset.ndim)]
        
        # handle all the tuple cases
        else:
            # index
            if isinstance(key[1], int):
                index = [key[1] + 1, key[1] + 2]
            elif isinstance(key[1], slice):
                index = [key[1].start + 1, key[1].stop + 2]
            else:
                raise KeyError('Batch index needs to be passed as int or slice.')
            
            # slices
            if len(key) == 2:
                slices = [[1, dataset.shape[i] + 1] for i in range(2, dataset.ndim)]
            else:  # more than 2
                slices = []
                for i, arg in enumerate(key[2:]):
                    if isinstance(arg, int):
                        slices.append([arg + 1, arg + 1])
                    elif isinstance(arg, slice):
                        slices.append([arg.start + 1 if arg.start is not None else 1, arg.stop + 1 if arg.stop is not None else dataset.shape[i + 1] + 1])
                    else:
                        raise KeyError('Slice needs to be passed as int or slice.')
                
                # check if we have all slices
                if len(slices) + 1 != dataset.ndim:
                    for i in range(len(slices) + 1, dataset.ndim):
                        slices.append([1, dataset.shape[i] + 1])
        
        # now, name, index and slices are set
        with self.backend.database() as db:
            # load the tensor
            arr = db.get_tensor(name, index[0], index[1], [s[0] for s in slices], [s[1] for s in slices])
        
        # TODO now we can transform to other libaries
        return arr

    def __setitem__(self, key: str, value: Union[List[list], np.ndarray]):
        # first make a numpy array from it
        if isinstance(value, list):
            value = np.asarray(value)

        # make at least 2D 
        if value.ndim == 1:
            value = value.reshape(1, -1)        
        
        # get the shape
        shape = value.shape

        # get the dim
        dim = value.ndim

        # check if this should be uplaoded chunk-wise
        if value.size > self.chunk_size:
            # figure out a good batch size
            batch_size = self.chunk_size // np.multiply(*value.shape[1:])
            if batch_size == 0:
                batch_size = 1
            
            # create the index over the batch to determine the offset on upload
            single_index = np.arange(0, value.shape[0], batch_size, dtype=int)
            batch_index = list(zip(single_index, single_index[1:].tolist() + [value.shape[0]]))
            
            # build the 
            batches = [(i * batch_size, value[up:low]) for i, (up, low) in enumerate(batch_index)]
        else:
            batches = [(0, value)]

        # connect
        with self.backend.database() as db:
            # insert the dataset
            dataset = db.insert_dataset(key, shape, dim)

            # make the iterator
            _iterator = tqdm(batches, desc=f'Uploading {key} [{len(batches)} batches of {batch_size}]') if not self.quiet else batches

            # insert the tensor
            for offset, batch in _iterator:
                db.insert_tensor(dataset.id, [tensor for tensor in batch], offset=offset)

    def __delitem__(self, key: str):
        with self.backend.database() as db:
            db.remove_dataset(key)
    
    def __contains__(self, key: str):
        # get the keys
        keys = self.keys()

        # check if key is in keys
        return key in keys
    
    def __len__(self):
        # get the keys
        keys = self.keys()

        # return the length
        return len(keys)

    def keys(self) -> List[str]:
        # get the keys from the database
        with self.backend.database() as db:
            keys = db.list_dataset_keys()
        
        # TODO: here caching could be added
        return keys



from typing import TYPE_CHECKING, List, Tuple
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy as np

if TYPE_CHECKING:
    from tensorage.session import BackendSession
    from tensorage.types import Dataset    


@dataclass
class BaseContext(ABC):
    backend: 'BackendSession' = field(repr=False)
    _anon_key: str = field(init=False, repr=False)

    @property
    def user_id(self) -> str:
        return self.backend._user.id

    @abstractmethod
    def get_dataset(self, key: str) -> 'Dataset':
        raise NotImplementedError

    @abstractmethod
    def get_tensor(self, key: str, index_low: int, index_up: int, slice_low: List[int], slice_up: List[int]) -> np.ndarray:
        raise NotImplementedError
    
    @abstractmethod
    def insert_dataset(self, key: str, shape: Tuple[int], dim: int, type: str, is_shared: bool) -> 'Dataset':
        raise NotImplementedError
    
    @abstractmethod
    def insert_tensor(self, data_id: int, data: List[np.ndarray], offset: int = 0) -> bool:
        raise NotImplementedError

    @abstractmethod
    def append_tensor(self, data_id: int, data: List[np.ndarray]) -> bool:
        raise NotImplementedError
    
    @abstractmethod
    def remove_dataset(self, key: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def list_dataset_keys(self) -> List[str]:
        raise NotImplementedError
    
    def __del__(self):
        self.backend.logout()


from typing import Optional
import os

from gotrue.types import AuthResponse

from .store import TensorStore
from .session import BackendSession

def login(email: str, password: str, backend_url: Optional[str] = None, backend_key: Optional[str] = None) -> TensorStore:
    # get the environment variables
    if backend_url is None:
        backend_url = os.getenv('SUPABASE_URL', 'http://localhost:8000')
    
    if backend_key is None:
        try:
            backend_key = os.environ['SUPABASE_KEY']
        except KeyError:
            raise RuntimeError('SUPABASE_KEY environment variable not set')
    
    # get a session
    session = BackendSession(email, password, backend_url, backend_key)

    # bind the session to the Store
    store = TensorStore(session)

    # return the store
    return store


def register(email: str, password: str, backend_url: Optional[str] = None, backend_key: Optional[str] = None) -> AuthResponse:
    # get the environment variables
    if backend_url is None:
        backend_url = os.getenv('SUPABASE_URL', 'http://localhost:8000')

    if backend_key is None:
        try:
            backend_key = os.environ['SUPABASE_KEY']
        except KeyError:
            raise RuntimeError('SUPABASE_KEY environment variable not set')
        
    # get a session
    session = BackendSession(None, None, backend_url, backend_key)

    # register
    response = session.register_by_mail(email, password)
    return response



from setuptools import setup, find_packages


def readme():
    with open('README.md') as f:
        return f.read()
    

def requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()


def version():
    loc = dict()
    with open('tensorage/__version__.py') as f:
        exec(f.read(), loc, loc)
        return loc['__version__']
    

setup(
    name='tensorage',
    version=version(),
    description='Storing tensors in a supabase backend',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Mirko Mälicke',
    author_email='mirko@hydrocode.de',
    install_requires=requirements(),
    packages=find_packages(),
)



SQL = """
-- datasets table
create table
public.datasets (
    id bigint generated by default as identity not null,
    key character varying not null,
    ndim smallint not null,
    shape int[] not null,
    created_at timestamp with time zone null default now(),
    user_id uuid null,
    is_public boolean not null default false,
    constraint datasets_pkey primary key (id),
    constraint datasets_key_user_id_key unique (key, user_id),
    constraint datasets_user_id_fkey foreign key (user_id) references users (id) on delete set null
) tablespace pg_default;

  -- RLS policy
ALTER TABLE public.datasets ENABLE ROW LEVEL SECURITY;
CREATE POLICY "Allow all actions to the record owner" ON "public"."datasets"
AS PERMISSIVE FOR ALL
TO authenticated
USING (is_public OR (auth.uid() = user_id))
WITH CHECK (is_public OR (auth.uid() = user_id));


-- tensor_float4 table
create table
public.tensors_float4 (
    data_id bigint not null,
    index bigint not null,
    tensor float4[] not null,
    user_id uuid not null,
    is_public boolean not null default false,
    constraint tensors_float4_pkey primary key (data_id, index, user_id),
    constraint tensors_float4_data_id_fkey foreign key (data_id) references datasets (id) on delete cascade,
    constraint tensors_float4_user_id_fkey foreign key (user_id) references users (id) on delete set null
) tablespace pg_default;

-- RLS policy
ALTER TABLE public.tensors_float4 ENABLE ROW LEVEL SECURITY;
CREATE POLICY "Allow all actions to the record owner" ON "public"."tensors_float4"
AS PERMISSIVE FOR ALL
TO authenticated
USING (is_public OR (auth.uid() = user_id))
WITH CHECK (is_public OR (auth.uid() = user_id));

-- create the slicing database function
CREATE OR REPLACE FUNCTION public.tensor_float4_slice(name character varying, index_low integer, index_up integer, slice_low integer[], slice_up integer[])
RETURNS table(tensor float4[])
AS
$$
DECLARE
  query_string text;
  i int;
BEGIN
  query_string := 'SELECT array_agg(tensors_float4.tensor';
  FOR i IN 1..array_length(slice_low, 1) LOOP
    query_string := query_string || '['|| slice_low[i] || ' : ' || slice_up[i] || ']';
  END LOOP;
  query_string := query_string || ') FROM tensors_float4
                   JOIN datasets ON datasets.id = tensors_float4.data_id
                   WHERE datasets.key = ' || quote_literal(name) || '
                   AND tensors_float4.index >= ' || index_low || '
                   AND tensors_float4.index < ' || index_up ;

  RETURN QUERY EXECUTE query_string;
END;
$$ language plpgsql;

-- add usage statistics views
create view
  public.user_usage_details as
select
  users.id,
  users.email,
  datasets.key as dataset,
  sum(pg_column_size(tensors_float4.tensor)::real) / 1024.0::double precision / 1024.0::double precision / 1024.0::double precision as total_size_gb,
  pg_size_pretty(sum(pg_column_size(tensors_float4.tensor))) as size
from
  tensors_float4
  join users on tensors_float4.user_id = users.id
  join datasets on datasets.id = tensors_float4.data_id
group by
  users.id,
  users.email,
  datasets.key;

create view
  public.user_usage_aggregate as
select
  users.id,
  users.email,
  sum(pg_column_size(tensors_float4.tensor)::real) / 1024.0::double precision / 1024.0::double precision / 1024.0::double precision as total_size_gb,
  pg_size_pretty(sum(pg_column_size(tensors_float4.tensor))) as size
from
  tensors_float4
  join users on tensors_float4.user_id = users.id
group by
  users.id,
  users.email;
"""